#ifndef __CAMERA_H
#define __CAMERA_H

#include "Vector3.h"
#include "AxisAlignedBox3.h"
#include "Plane3.h"


namespace CHCDemoEngine 
{

class Matrix4x4;
class Polyhedron;

/** Class representing a frustum.
*/
class Frustum
{
	friend class Camera;

public:

	Frustum() {};

	Frustum(const Matrix4x4 &trafo);

	enum { LEFT_PLANE, RIGHT_PLANE, BOTTOM_PLANE, TOP_PLANE, NEAR_PLANE, FAR_PLANE};

	/** Resizes the current frustum so that it fully encloses the given polyhedron
	*/
	void EnclosePolyhedron(const Polyhedron &polyhedron);

	void ExtractTransformation(Matrix4x4 &m) const;


	/// the 6 clip planes
	Plane3 mClipPlanes[6];
};


/** Classs representing a camera.
*/
class Camera
{
	friend class ShadowMap;

public:
	/** Default constructor.
	*/
	Camera();
	/** Camera taking the image width and height and the field of view.
	*/
	Camera(float aspect, float fieldOfView = 90.f);
	/** Sets the current camera position.
	*/
	void SetPosition(const Vector3 &pos);
	/** See set.
	*/
	inline Vector3 GetPosition() const { return mPosition; }
	/** Returns view direction.
	*/
	Vector3 GetDirection() const;
	/** Returns up vector.
	*/
	Vector3 GetUpVector() const;
	/** Returns right vector.
	*/
	Vector3 GetRightVector() const;

	Vector3 GetBaseDirection() const;
	Vector3 GetBaseUpVector() const;
	Vector3 GetBaseRightVector() const;
	/** Returns the field of view.
	*/
	inline float GetFov() const { return mFOVy; }
	/** Returns the aspect ratio.
	*/
	inline float GetAspect() const { return mAspect; }
	/** Sets up viewing matrices in for opengl rendering
	*/
	void SetupCameraView();
	/** Returns the current projection matrix.
	*/
	void GetProjectionMatrix(Matrix4x4 &mat) const;
	/** Returns the current model view matrix.
	*/
	void GetModelViewMatrix(Matrix4x4 &mat) const;
	/** Returns the view orientation (the model view matrix without the translation)
	*/
	void GetViewOrientationMatrix(Matrix4x4 &mat) const;
	/** Calculates a frustum from the projection and the modelview matrix.
	*/
	void CalcFrustum(Frustum &frustum);
	/** Computes the extremal points of this frustum.
		If farthestVisibleDistance is nearer than the far plane, 
		it is used to define the far plane instead.
	*/
	void ComputePoints(Vector3 &ftl, Vector3 &ftr, Vector3 &fbl, Vector3 &fbr,
		               Vector3 &ntl, Vector3 &ntr, Vector3 &nbl, Vector3 &nbr,
					   float farthestVisibleDistance = 1e25f) const;
	/** Returns the near plane.
	*/
	inline float GetNear() const { return mNear; }
	/** Returns the far plane.
	*/
	inline float GetFar() const { return mFar; }
	/** Sets the near plane
	*/
	void SetNear(float nearDist);
	/** Sets the far plane.
	*/
	void SetFar(float farDist);
	/** Sets this camera to orthographic projection.
	*/
	//void SetOrtho(bool ortho);
	/** Set yaw (rotation around vertical axis.
	*/
	void Yaw(float angle);
	/** Set pitch.
	*/
	void Pitch(float angle);
	/** Returns pitch.
	*/
	float GetPitch() const { return mPitch; }
	/** Returns yaw.
	*/
	float GetYaw() const { return mYaw; }
	/** Resets pitch and yaw.
	*/
	void ResetPitchAndYaw() { mPitch = 0; mYaw = 0; }
	/** Sets the view direction.
	*/
	void SetDirection(const Vector3 &direction);
	/** Returns frustum as polyhedron. 
	*/
	Polyhedron *ComputeFrustum(float farthestVisibleDistance = 1e25f) const;
	/** Sets up projection matrix in OpenGl.
	*/
	void SetupProjection();
	/** Sets up view + projection matrix in OpenGl.
	*/
	void SetupViewProjection();


protected:

	void Precompute(const Vector3 &direction);

	void CalculateFromPitchAndYaw();

	void ComputePointsInternal(Vector3 &ftl, Vector3 &ftr, Vector3 &fbl, Vector3 &fbr,
							   Vector3 &ntl, Vector3 &ntr, Vector3 &nbl, Vector3 &nbr,
							   const Vector3 &view, const Vector3 &right, const Vector3 &up,
							   const Vector3 &pos,
							   float farthestVisibleDistance = 1e25f) const;

	////////////////
	//-- members

	float mFOVy;
	float mNear;
	float mFar;
	
	/// if this camera is orthgraphic or perspective
	//bool mIsOrtho;

	Matrix4x4 mBaseOrientation;
	Matrix4x4 mViewOrientation;

	float mPitch;
	float mYaw;

	Vector3 mPosition;

	float mAspect;
};

}
#endif
